/***
 * Excerpted from "Modern Systems Programming with Scala Native",
 * published by The Pragmatic Bookshelf.
 * Copyrights apply to this code. It may not be used to create training material,
 * courses, books, articles, and the like. Contact us if you are in doubt.
 * We make no guarantees that this code is fit for any purpose.
 * Visit http://www.pragmaticprogrammer.com/titles/rwscala for more book information.
***/
import scala.scalanative.unsafe._
import scala.scalanative.libc._
import stdlib._
import string._
import stdio._

final case class WrappedArray[T](var data:Ptr[T], var used:Int, var capacity:Int)

object main {
  type NGramData = CStruct4[CString, Int, Int, Int]

  def makeWrappedArray(size:Int):WrappedArray[NGramData] = {
    val data = malloc(size * sizeof[NGramData]).asInstanceOf[Ptr[NGramData]]
    return WrappedArray[NGramData](data, 0, size)
  }

  def growWrappedArray(array:WrappedArray[NGramData], size:Int):Unit = {
    val new_capacity = array.capacity + size
    val new_size = new_capacity * sizeof[NGramData]
    val new_data = realloc(array.data.asInstanceOf[Ptr[Byte]], new_size)
    array.data = new_data.asInstanceOf[Ptr[NGramData]]
    array.capacity = new_capacity
  }

  val by_count = new CFuncPtr2[Ptr[Byte],Ptr[Byte],Int] {
    def apply(p1:Ptr[Byte], p2:Ptr[Byte]):Int = {
      val ngram_ptr_1 = p1.asInstanceOf[Ptr[NGramData]]
      val ngram_ptr_2 = p2.asInstanceOf[Ptr[NGramData]]
      val count_1 = ngram_ptr_1._2
      val count_2 = ngram_ptr_2._2
      return count_2 - count_1
    }
  }

val by_count_naive = new CFuncPtr2[Ptr[Byte],Ptr[Byte],Int] {
  def apply(p1:Ptr[Byte], p2:Ptr[Byte]):Int = {
    val ngram_ptr_1 = p1.asInstanceOf[Ptr[NGramData]]
    val ngram_ptr_2 = p2.asInstanceOf[Ptr[NGramData]]
    val count_1 = ngram_ptr_1._2
    val count_2 = ngram_ptr_2._2
    if (count_1 > count_2) {
      return -1
    }
    else if (count_1 == count_2) {
      return 0
    }
    else {
      return 1
    }
  }
}

def main(args:Array[String]):Unit = {
  val block_size = 65536 * 16
  val line_buffer = malloc(1024)

  var array = makeWrappedArray(block_size)

  val read_start = System.currentTimeMillis()
  while (stdio.fgets(line_buffer, 1023, stdin) != null) {
    if (array.used == array.capacity) {
        growWrappedArray(array, block_size)
    }
    parseLine(line_buffer, array.data + array.used)
    array.used += 1
  }

  val read_elapsed = System.currentTimeMillis() - read_start
  stdio.fprintf(stdio.stderr, c"reached EOF after %d lines in %d ms\n",
                array.used, read_elapsed)

  val sort_start = System.currentTimeMillis()
  qsort.qsort(array.data.asInstanceOf[Ptr[Byte]], array.used,
    sizeof[NGramData], by_count)
  val sort_elapsed = System.currentTimeMillis() - sort_start
  stdio.printf(c"sorting done in %d ms\n", sort_elapsed)

  val to_show = if (array.used <= 20) array.used else 20

  for (i <- 0 until to_show) {
    stdio.printf(c"word %d: %s %d\n", i, (array.data + i)._1,
                 (array.data + i)._2)
  }
  stdio.printf(c"done.\n")
}

val temp_word = malloc(1024)
def parseLine(line_buffer:Ptr[Byte], data:Ptr[NGramData]):Unit = {
  val count = data.at2
  val year = data.at3
  val doc_count = data.at4

  val sscanf_result = stdio.sscanf(line_buffer, c"%1023s %d %d %d\n",
                                   temp_word, year, count, doc_count)
  if (sscanf_result < 4) {
    throw new Exception("input error")
  }
  val word_length = strlen(temp_word)
  val new_string = malloc(word_length + 1)
  strncpy(new_string, temp_word, word_length + 1)
  data._1 = new_string
}

def freeArray(array:Ptr[NGramData], size:Int):Unit = {
  for (i <- 0 until size) {
    val item = array + i
    stdlib.free( item._1 )
  }
  stdlib.free(array.asInstanceOf[Ptr[Byte]])
}

def parseLine_naive(fd:Ptr[stdio.FILE], array:Ptr[NGramData]):Unit = {
  ??? // TODO
}
}

@extern
object qsort {
  def qsort(data:Ptr[Byte],
          num:Int,
          size:Long,
          comparator:CFuncPtr2[Ptr[Byte], Ptr[Byte], Int]):Unit = extern

}
